""" Main script.
"""
import copy
import functools
#import time
#import sys
import glob
import hashlib
import inspect
import os
import pickle
from typing import Any, Callable, List, Optional, Tuple, Union

import six

from . import _utils
from .config import Config
from .myfilelock import FileLock, Timeout

_DEBUG = False
NOCACHE, CACHE, RECACHE, READONLY = [0, 1, 2, 3]


def _print(*args, **kwargs):
    if _DEBUG:
        print(*args, **kwargs)


def get_persist_dir_from_paths(base_dir: str, file_path: str, project_path: str):
    """Infers the directory to store all cache data, basing on the project_path,
     file_path (which should be inside project_path) and the base cache directory base_dir.

    Args:
        base_dir (str): Directory for all cache for project_path.
        file_path (str): The path to the .py file we are interested in.
        project_path (str): Where the project lives (i.e. what os.getcwd() returns).

    Returns:
        str: The directory to store cache for functions in file_path.
    """
    file_path = os.path.normpath(os.path.abspath(file_path))
    assert file_path.endswith(".py")
    file_path = os.path.splitext(file_path)[0]

    if project_path is None:
        base_dir = os.path.join(base_dir, '.hashed')
        _utils.make_dir_if_necessary(base_dir)
        meta_file = os.path.join(base_dir, 'mapping.txt')
        pid = _utils.retrieve_id(meta_file, file_path, sep='||')
        persist_dir = os.path.join(
            base_dir, f"{os.path.basename(file_path)}-{pid}")
    else:
        project_path = os.path.normpath(os.path.abspath(project_path))
        assert project_path in file_path, \
            f"Expect file_path to be a sub-directory of project_path,'\
            ' but got {file_path} and {project_path}"
        persist_dir = file_path.replace(
            project_path, os.path.normpath(base_dir))
    return os.path.normpath(persist_dir)


def _hash(k):
    return int(hashlib.md5(pickle.dumps(k, protocol=3)).hexdigest(), 16)


def _persist_rw_curr_results(cache_path, write_key=None, write_val=None, *, lock_path):
    if lock_path is None:
        lock_path = cache_path  # lock at call level
    with FileLock(lock_path):
        try:
            res = _utils.read_pickle(cache_path)
        except Exception as err:
            if os.path.exists(cache_path):
                print(f"Error: {err}. Re-creating a new cache")
            res = {}
            _utils.to_pickle(res, cache_path)
        if write_key is not None:
            res[write_key] = write_val
            _utils.to_pickle(res, cache_path)
        return res


def _persist_write(cache_path, key, closure_func: Callable[[], Any], alt_roots, *, lock_path):
    if alt_roots is not None:
        # try to look up in alterantive root cache paths..
        # If can't find anything, then save to the original path.
        # This is like merging cache.
        # assert _PERSIST_PATH in cache_path, "How did you generate this cache path??"
        raise NotImplementedError()  # Should probably replace before passing in alt_roots
    else:
        val = closure_func()
    try:
        _persist_rw_curr_results(cache_path, key, val, lock_path=lock_path)
    except Timeout as err:
        raise err
    return val


def _persist_write_if_necessary(cache_path, key, closure_func: Callable[[], Any],
                                readonly=False, alt_roots=None, *, lock_path):
    try:
        _print(
            f"persist_to_disk: {cache_path} exists? : {os.path.isfile(cache_path)}.")
        res = _persist_rw_curr_results(cache_path, lock_path=lock_path)
        _print(
            f"persist_to_disk: Looking up {key} in {cache_path}({res.keys()}): {key in res}.")
        if key in res:
            return res[key]
    except Timeout as err:
        raise err
    assert not readonly, "In readonly mode, but there is no existing cache."
    return _persist_write(cache_path, key, closure_func, alt_roots=alt_roots, lock_path=lock_path)


# test input d={"model": {"1": {"2": 3, '2a': 4}}, 'a': 2}
def _expand_dict_recursive(d):
    old_d = d
    d = {}
    for k, v in old_d.items():
        if isinstance(v, dict):
            v = _expand_dict_recursive(v)
            for kk, vv in v.items():
                d[f"{k}|{kk}"] = vv
        else:
            d[k] = v
    return d


def _get_full_kwargs_noargs(func, args1, kwargs1):
    args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = inspect.getfullargspec(
        func)
    assert varargs is None, "Does not support functions with *args."
    full_kwargs = {}
    # default values
    if kwonlydefaults is not None:
        full_kwargs.update(kwonlydefaults)
    if defaults is not None:
        full_kwargs.update(dict(zip(args[-len(defaults):], defaults)))
    # input values
    for argi, argv in enumerate(args1):
        full_kwargs[args[argi]] = argv
    full_kwargs.update(kwargs1)
    return full_kwargs


def _clean_kwargs(full_kwargs, skip_kwargs, expand_dict_kwargs):
    for k in skip_kwargs:
        full_kwargs.pop(k, None)
    if isinstance(expand_dict_kwargs, str) and expand_dict_kwargs == 'all':
        full_kwargs = _expand_dict_recursive(full_kwargs.copy())
    else:
        for k in expand_dict_kwargs:
            if k in full_kwargs.keys():
                d = full_kwargs.pop(k)
                full_kwargs.update({f"{k}|{kk}": v for kk, v in d.items()})
    return full_kwargs


def _get_hashed_path_and_key(cache_dir, full_kwargs, hashsize, groupby):
    for k in groupby:
        if isinstance(k, tuple):
            dirname = "$$".join([str(full_kwargs.pop(kk)) for kk in k])
        else:
            dirname = str(full_kwargs.pop(k))
        cache_dir = os.path.join(cache_dir, dirname)
    _utils.make_dir_if_necessary(cache_dir)
    key = tuple(sorted(six.iteritems(full_kwargs), key=lambda x: x[0]))
    hashed_path = os.path.join(cache_dir, f"{_hash(key) % hashsize}.pkl")
    return hashed_path, key


def _get_lock_path(call_cache_path, config: Config) -> str:
    lock_granularity = config.config['lock_granularity']
    if lock_granularity == 'call':
        return call_cache_path
    if lock_granularity == 'func':
        return os.path.join(os.path.dirname(call_cache_path), 'func_persist_lock')
    assert lock_granularity == 'global'
    return os.path.join(config.get_project_persist_path(), 'global_persist_lock')


class Persister():
    """Base class that does all the heavy-lifting.
    """

    @classmethod
    def _check_arguments(cls, config: Config, freq, skip_kwargs, hashsize, switch_kwarg,
                         expand_dict_kwargs, groupby, cache):
        if isinstance(expand_dict_kwargs, str) and expand_dict_kwargs == 'all':
            expand_dict_kwargs = []
        special_kwargs = set(groupby + expand_dict_kwargs + skip_kwargs)
        assert len(special_kwargs) == len(groupby) + \
            len(expand_dict_kwargs) + len(skip_kwargs)
        assert switch_kwarg not in special_kwargs

    def __init__(self, func: Callable, config: Config,
                 freq=None, hashsize: int = None,
                 skip_kwargs: List[str] = None, expand_dict_kwargs: Union[List[str], str] = None,
                 groupby: List[str] = None,
                 switch_kwarg: str = 'cache_switch', cache: int = None):
        functools.update_wrapper(self, func)
        self.__defaults__ = six.get_function_defaults(func)
        if skip_kwargs is None:
            skip_kwargs = []
        if expand_dict_kwargs is None:
            expand_dict_kwargs = []
        if groupby is None:
            groupby = []

        Persister._check_arguments(
            config, None, skip_kwargs, hashsize, switch_kwarg, expand_dict_kwargs, groupby, cache)

        assert inspect.getfullargspec(
            func)[1] is None, "Does not support functions with *args."
        # self.__dict__.update(func.__dict__)
        # infodict = get_info(func)

        # workspace config
        self.config = config

        # current function settings
        self.hashsize = hashsize or config.get_hashsize()
        self.freq = freq
        self.skip_kwargs = skip_kwargs
        self.switch_kwarg = switch_kwarg  # 0 is not cache, 1 is cache, 2 is recache
        self.expand_dict_kwargs = expand_dict_kwargs
        self.groupby = groupby

        # Get the cache_dir straight
        self.cache_dir = get_persist_dir_from_paths(
            config.get_project_persist_path(),
            inspect.getsourcefile(func),
            config.get_project_path()
        )
        self.cache_dir = os.path.join(self.cache_dir, self.__name__)
        assert '__main__' not in self.cache_dir
        _utils.make_dir_if_necessary(self.cache_dir)
        assert freq is None, "Not implemented yet"
        self.alt_roots = config.get_alternative_roots()

        self.cache = cache

    def __call__(self, *args, **kwargs):
        kwargs = copy.deepcopy(kwargs)
        curr_cache_switch = int(kwargs.pop(self.switch_kwarg, CACHE))

        def closure():
            return self.__wrapped__(*args, **kwargs)

        full_kwargs = _get_full_kwargs_noargs(self.__wrapped__, args, kwargs)
        cache_switch = self.cache if self.cache is not None else curr_cache_switch
        if cache_switch == NOCACHE:
            return self.__wrapped__(**full_kwargs)
        _cleaned = _clean_kwargs(
            full_kwargs, self.skip_kwargs, self.expand_dict_kwargs)
        hashed_path, key = _get_hashed_path_and_key(
            self.cache_dir, _cleaned, self.hashsize, self.groupby)
        lock_path = _get_lock_path(hashed_path, self.config)

        if cache_switch == RECACHE:
            return _persist_write(hashed_path, key, closure, alt_roots=None, lock_path=lock_path)
        return _persist_write_if_necessary(hashed_path, key, closure,
                                           readonly=cache_switch == READONLY,
                                           alt_roots=self.alt_roots, lock_path=lock_path)

    def clear(self):
        """clean all the cache for self.__wrapped__
        """
        files = glob.glob(f'{self.cache_dir}/*')
        for f in files:
            os.remove(f)


# function version =====================
def persist_func_version(func, config: Config, **kwargs):
    """wrapper so that the func-to-persist can be pickled (e.g. in multiprocessing)
    """
    obj = Persister(func, config, **kwargs)

    @functools.wraps(func)
    def inner(*args, **kwargs):
        return obj(*args, **kwargs)
    return inner

# ===========================Manual Cache


def _get_caller_cache_path(config: Config, caller_=None, make_if_necessary=False):
    cache_dir = get_persist_dir_from_paths(
        config.get_project_persist_path(),
        caller_.filename,
        config.get_project_path()
    )
    if 'self' in caller_[0].f_locals:
        _caller_func_name = f"{caller_[0].f_locals['self'].__class__.__name__}.{caller_.function}"
    else:
        _caller_func_name = caller_.function
    cache_dir = os.path.join(cache_dir, _caller_func_name)
    assert '__main__' not in cache_dir
    if make_if_necessary:
        _utils.make_dir_if_necessary(cache_dir)
    return cache_dir


def _manual_cache_infer_path(key, obj, write, config: Config, caller_):
    cache_dir = _get_caller_cache_path(config, caller_, True)
    cache_path = os.path.join(cache_dir, key)
    if write:
        _utils.to_pickle(obj, cache_path)
    else:
        if os.path.exists(cache_path):
            return _utils.read_pickle(cache_path)
        return None



""" Util functions
"""
import pickle
import os

from .myfilelock import FileLock

PICKLE_PROTOCOL = 4


def dump(obj, file, **kwargs):
    """like pickle.dump but fix a default protocol for compatibility.
    """
    kwargs.setdefault('protocol', PICKLE_PROTOCOL)
    return pickle.dump(obj, file, **kwargs)


def to_pickle(obj, filepath, **kwargs):
    """Like to_pickle in pandas, but avoids pandas dependency.
    """
    with open(filepath, 'wb') as fout:
        dump(obj, fout, **kwargs)


def read_pickle(filepath, **kwargs):
    """Like read_pickle in pandas, but avoids pandas dependency.
    """
    with open(filepath, 'rb') as fin:
        return pickle.load(fin, **kwargs)


def make_dir_if_necessary(dirname, max_depth=3):
    """Check if a directory exists. If not make it with lock.
    Will create nested directories (up to depth=max_depth).

    Args:
        dirname (_type_):
            Target directory.
        max_depth (int, optional):
            Maximum depth for nested creation.
            Defaults to 3.
    """
    if os.path.isdir(dirname):
        return
    assert max_depth >= 0, f"Cannot make too many nested directories. Something could be wrong!, dirname={dirname}"
    if not os.path.isdir(os.path.dirname(dirname)):
        make_dir_if_necessary(os.path.dirname(dirname), max_depth - 1)
    fl = FileLock(dirname)
    with fl:
        print(f"{dirname} does not exist. Creating it for persist_to_disk")
        os.makedirs(dirname)
    return


def retrieve_id(meta_file, key, sep='||'):
    """An internal helper to retrieve a mapped id (and create one if necessary).
    """
    with FileLock(meta_file):
        curr_dict, lines = dict(), []
        if os.path.isfile(meta_file):
            with open(meta_file, 'r', encoding='utf-8') as fin:
                lines = fin.readlines()
                curr_dict = dict([line.strip().split(sep) for line in lines])
        if key not in curr_dict:
            curr_dict[key] = pid = str(len(curr_dict)+1)
            with open(meta_file, 'a', encoding='utf-8') as fout:
                fout.write(f"{key}{sep}{pid}\n")
    return curr_dict[key]



import os
from pathlib import Path

from setuptools import find_packages, setup


def create_setting_folder():
    """Create the setting folder
    """
    SETTING_PATH = os.path.join(Path.home(), '.persist_to_disk')
    if not os.path.isdir(SETTING_PATH):
        os.makedirs(SETTING_PATH)


with open('README.md', encoding='utf-8') as readme_file:
    README = readme_file.read()

with open('HISTORY.md', encoding='utf-8') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='persist_to_disk',
    version='0.0.4',
    description='Persist expensive operations on disk.',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='Zhen Lin',
    author_email='zhenlin4@illinois.edu',
    keywords=['Cache', 'Persist'],
    url='https://github.com/zlin7/python-persist_to_disk',
    # download_url='https://pypi.org/project/elastictools/'
)

install_requires = [
    'filelock==3.9.0'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)
    create_setting_folder()



from filelock import Timeout, FileLock as RawFileLock
assert Timeout is not None


class FileLock(object):
    """A wrapper for filelock.FileLock
    """

    def __init__(self, protected_file_path, timeout=5):
        """ Prepare the file locker. Specify the file to lock and optionally
                the maximum timeout and the delay between each attempt to lock.
        """
        self.lock_path = protected_file_path + ".lock"
        self.lock = RawFileLock(self.lock_path, timeout=timeout)

    def __enter__(self):
        return self.lock.__enter__()

    def __exit__(self, *args, **kwargs):
        return self.lock.__exit__(*args, **kwargs)

    def __del__(self):
        return self.lock.__del__()



"""Main functions for persist_to_disk.
"""
import inspect
import os
from typing import Any, Callable, List, Optional, Tuple, Union

from . import persister
from .config import Config
from .persister import (CACHE, NOCACHE, READONLY, RECACHE, Persister,
                        persist_func_version)

# Global config so user could set the root directory for persist ops
config = Config()


def persistf(freq=None, hashsize: int = None,
             skip_kwargs: List[str] = None, expand_dict_kwargs: Union[List[str], str] = None,
             groupby: List[str] = None,
             switch_kwarg: str = 'cache_switch', cache: int = None):
    """Base decorator that does all the heavy-lifting for caching, taking additional arguments.

    Args:
        freq (_type_, optional): unused for now. Defaults to None.
        hashsize (int, optional): Calls/Inputs to func are hashed into *hashsize* buckets.
            Defaults to what's set in config.
        skip_kwargs (List[str], optional): these kwargs are ignored (e.g. gpu_id, verbose, ...).
            Defaults to empty list.
        expand_dict_kwargs (Union[List[str], str], optional): arguments to expand.
            Applies to arguments to *func* that are dictionaries.
            If str, it must be `all` which means all dictionaries are recursively expanded.
            Note that any calls with un-expanded dictionary arguments will fail.
            Defaults to None.
        groupby (List[str], optional): Create several levels of cache.
            For example, if groupby=['dataset'], then calls with each
            dataset will be stored in a separate folder.
            Defaults to None.
        switch_kwarg (str, optional):
            A switch to turn on/off the caching mechanism.
            Takes value in {NOCACHE, CACHE, RECACHE, READONLY}
            Defaults to 'cache_switch'.
        cache (int, optional):
            Same as switch_kwarg but this is a function-level setting (not call level).
            Useful for recaching/debugging purposes.
            Defaults to None (equivalent to CACHE).
    """
    def _decorator(func):
        return persist_func_version(func, config,
                                    freq=freq, hashsize=hashsize,
                                    skip_kwargs=skip_kwargs, expand_dict_kwargs=expand_dict_kwargs,
                                    groupby=groupby, switch_kwarg=switch_kwarg, cache=cache)
    return _decorator


def persist(freq=None, hashsize: int = None,
            skip_kwargs: List[str] = None, expand_dict_kwargs: Union[List[str], str] = None,
            groupby: List[str] = None,
            switch_kwarg: str = 'cache_switch', cache: int = None):
    """Base decorator that does all the heavy-lifting for caching, taking additional arguments.

    Args:
        freq (_type_, optional): unused for now. Defaults to None.
        hashsize (int, optional): Calls/Inputs to func are hashed into *hashsize* buckets.
            Defaults to what's set in config.
        skip_kwargs (List[str], optional): these kwargs are ignored (e.g. gpu_id, verbose, ...).
            Defaults to empty list.
        expand_dict_kwargs (Union[List[str], str], optional): arguments to expand.
            Applies to arguments to *func* that are dictionaries.
            If str, it must be `all` which means all dictionaries are recursively expanded.
            Note that any calls with un-expanded dictionary arguments will fail.
            Defaults to None.
        groupby (List[str], optional): Create several levels of cache.
            For example, if groupby=['dataset'], then calls with each
            dataset will be stored in a separate folder.
            Defaults to None.
        switch_kwarg (str, optional):
            A switch to turn on/off the caching mechanism.
            Takes value in {NOCACHE, CACHE, RECACHE, READONLY}
            Defaults to 'cache_switch'.
        cache (int, optional):
            Same as switch_kwarg but this is a function-level setting (not call level).
            Useful for recaching/debugging purposes.
            Defaults to None (equivalent to CACHE).
    """
    return lambda func: Persister(func, config,
                                  freq=freq, hashsize=hashsize,
                                  skip_kwargs=skip_kwargs, expand_dict_kwargs=expand_dict_kwargs,
                                  groupby=groupby, switch_kwarg=switch_kwarg, cache=cache)


def clear_locks(clear=False):
    """This function clears ALL locks for your project, if any.
    Such locks could be created in multi-process usage of persist_to_disk.
    Please only use it when you are sure no process is still using these locks to access files.

    Args:
        clear (bool, optional):
            If clear=False, only prints the dead locks.
            If True, actually delete them.
            Defaults to False.
    """
    for root, dirs, files in os.walk(config.get_project_persist_path(), topdown=False):
        for name in files:
            if name.endswith(".lock"):
                print(os.path.join(root, name))
                if clear:
                    os.remove(os.path.join(root, name))
        continue
        for name in dirs:
            print(os.path.join(root, name))


def get_caller_cache_path(make_if_necessary=True):
    """infer the cache path for the caller, for manual cache.

    Args:
        make_if_necessary (bool, optional):
            Make the directory if it does not exist.
            Defaults to True.

    Returns:
        str: default path to save the cache
    """
    return persister._get_caller_cache_path(config, inspect.stack()[1], make_if_necessary)


def manual_cache(key: str, obj: Any = None, write: bool = False) -> Any:
    """Manual cache helper.
    Each function gets a directory to store all results.
    Each result is saved with *key* as the filename.
    If *write*, writes *obj* with key=*key*.

    WARNING: This function is not multiprocess-safe.
        This is because each file should contain only 1 cache result.

    Args:
        key (str):
            key of cache.
        obj (Any, optional):
            Result to cache. Defaults to None.
        write (bool, optional):
            Write or read the cache. Defaults to False.

    Returns:
        Any: cached result when *write*, else None.
    """
    return persister._manual_cache_infer_path(key, obj, write, config, inspect.stack()[1])


__all__ = [
    'config',
    'persist',
    'clear_locks',
    'persistf',
    'get_caller_cache_path',
    'manual_cache'
]

__version__ = "0.0.4"
__author__ = 'Zhen Lin'
__credits__ = ''



import os
from pathlib import Path
import configparser

from .myfilelock import FileLock
from . import _utils

SETTING_PATH = os.path.join(Path.home(), '.persist_to_disk')
_utils.make_dir_if_necessary(SETTING_PATH)
DEFAULT_PERSIST_PATH = os.path.join(SETTING_PATH, 'cache')
# DEFAULT_PERSIST_PATH = 'C:/Users/zhen7/Desktop/gitRes/persist_to_disk/.persist_to_disk'
CONFIG_PATH = os.path.join(SETTING_PATH, 'config.ini')


def _read_project_pid(project_path, sep='||'):
    assert sep not in project_path
    project_path = os.path.normpath(project_path)
    meta_file = os.path.join(SETTING_PATH, 'project_to_pids.txt')
    return _utils.retrieve_id(meta_file, project_path, sep=sep)


def _record_project_persist_path(persist_path, pid, sep='||'):
    assert sep not in persist_path
    meta_file = os.path.join(SETTING_PATH, 'pid_to_persist_dirs.txt')
    with FileLock(meta_file):
        curr_dict, lines = {}, []
        if os.path.isfile(meta_file):
            with open(meta_file, 'r', encoding='utf-8') as fin:
                lines = fin.readlines()
            for line in lines:
                line = line.strip().split(sep)
                curr_dict[line[0]] = set(line[1:])
        if pid not in curr_dict:
            curr_dict[pid] = set([persist_path])
            with open(meta_file, 'a', encoding='utf-8') as fout:
                fout.write(
                    f"{pid}{sep}{sep.join(list(curr_dict[pid]))}\n")
        elif persist_path not in curr_dict[pid]:
            for i, line in enumerate(lines):
                if line.strip().split(sep)[0] == pid:
                    lines[i] = f"{line.strip()}{sep}{persist_path}"
                with open(meta_file, 'w', encoding='utf-8') as fout:
                    fout.writelines(lines)


class Config(dict):
    """Global config for this project.
    """

    def __init__(self) -> None:
        global_config = configparser.ConfigParser()
        if not os.path.isfile(CONFIG_PATH):
            global_config.add_section('global_settings')
            global_config['global_settings']['persist_path'] = DEFAULT_PERSIST_PATH
            global_config['global_settings']['hashsize'] = '500'

            # global, func, call
            global_config['global_settings']['lock_granularity'] = 'func'

            # List to add:
            # 1. Whether to automatically clear a folder if hashsize changes
        else:
            global_config.read(CONFIG_PATH)
        self.global_config = global_config
        # worksapce_config
        self.config = {'persist_path': None,
                       'project_path': os.path.normpath(os.getcwd())}
        self.set_persist_path(
                self.global_config['global_settings']['persist_path'])

        self._private_config = {'_persist_path': {}}
        for key in ['hashsize', 'lock_granularity']:
            self.config[key] = self.global_config['global_settings'][key]
        assert self.config['lock_granularity'] in {"call", "func", "global"}

    def generate_config(self):
        with open(CONFIG_PATH, 'w', encoding='utf-8') as fout:
            self.global_config.write(fout)
        print(f"Settings written to {CONFIG_PATH}")

    def set_project_path(self, path: str):
        self.config['project_path'] = os.path.normpath(os.path.abspath(path))
        return self.config['project_path']

    def set_persist_path(self, path):
        self.config['persist_path'] = os.path.normpath(os.path.abspath(path))
        return self.config['persist_path']

    def set_hashsize(self, hashsize=500):
        self.config['hashsize'] = hashsize
        return hashsize

    def set_alternative_readonly_persist_paths(self, paths):
        raise NotImplementedError()

    def get_alternative_roots(self):
        return None

    def _compute_and_save_actual_persist_path(self):
        persist_path = self.get_persist_path()
        project_path = self.get_project_path()

        if (persist_path, project_path) not in self._private_config['_persist_path']:
            pid = _read_project_pid(project_path)
            final_path = os.path.join(persist_path, f"{os.path.basename(project_path)}-{pid}")
            _record_project_persist_path(final_path, pid)
            _utils.make_dir_if_necessary(final_path)
            self._private_config['_persist_path'][(persist_path, project_path)] = final_path
        return self._private_config['_persist_path'][(persist_path, project_path)]

    def get_persist_path(self):
        return self.config['persist_path']

    def get_project_persist_path(self):
        return self._compute_and_save_actual_persist_path()

    def get_project_path(self):
        return self.config['project_path']

    def get_hashsize(self):
        return int(self.config['hashsize'])
